"""Constants module."""

from cmem_plugin_base.dataintegration.parameter.code import SparqlCode

POLICY_TEMPLATE = """
grant {{
permission java.util.PropertyPermission "user.dir", "read";
permission java.util.PropertyPermission "http://jena.hpl.hp.com/2004/07/feature/noSecurity", "read";
permission java.util.PropertyPermission "http://jena.hpl.hp.com/2004/07/feature/noCharset", "read";
permission java.util.PropertyPermission "http://jena.hpl.hp.com/2004/08/LocationMap", "read";
permission java.util.PropertyPermission "LocationMap", "read";
permission java.util.PropertyPermission "org.apache.jena.tdb.settings", "read";
permission java.util.PropertyPermission "tdb:settings", "read";
permission java.util.PropertyPermission "sun.arch.data.model", "read";
permission java.util.PropertyPermission "*", "read,write";
permission java.lang.RuntimePermission "getenv.TURN_OFF_LR_LOOP_ENTRY_BRANCH_OPT";
permission java.io.FilePermission "location-mapping.ttl", "read";
permission java.io.FilePermission "etc/location-mapping.ttl", "read";
permission java.io.FilePermission "etc/location-mapping.rdf", "read";
permission java.io.FilePermission "location-mapping.rdf", "read";

permission java.io.FilePermission "{}", "read";
permission java.io.FilePermission "{}", "read";

}};
"""

DEFAULT_SPARQL = SparqlCode("""PREFIX fx: <http://sparql.xyz/facade-x/ns/>
CONSTRUCT { ?s ?p ?o }
WHERE
{
  SERVICE <x-sparql-anything:{{resource_file}}> {
    fx:properties fx:root "https://example.org/document" .
    fx:properties fx:blank-nodes false .
    ?s  ?p  ?o .
  }
}""")

QUERY_PARAMETER_DESCRIPTION = f"""
Query to run with sparql-anything engine.

Note: resource_file place holder will be replaced with selected resource file.

Example
```
{DEFAULT_SPARQL}
```
For more information about sparql-anything please refer: https://sparql-anything.readthedocs.io/stable/
"""

SPARQL_ANYTHING_ERROR_PATTERN = "io.github.sparqlanything.cli.SPARQLAnything"

DOCUMENTATION = """[SPARQL Anything](https://sparql-anything.cc/) is a system for Semantic Web
re-engineering that allows users to ... query anything with SPARQL.

This workflow task allows for execution of SPARQL queries against the simplistic Facade-X
meta-model of SPARQL Anything.
Facade-X provides a homogeneous view over heterogeneous data sources and support multiple formats.
It allows to extract data from project files by using SPARQL Construct queries.
In order to reference, the file, you need to use the `{{resource_file}}` placeholder.
"""



"""cmem-plugin-sparql-anything"""

import logging

import requests

from cmem_plugin_sparql_anything import utils

try:
    if not utils.has_jar():
        utils.download_sparql_anything()
except requests.exceptions.RequestException:
    logging.info("Failed to download sparql anything zip file")
    raise



"""SPARQL Anything jar bin module"""

import os
from pathlib import Path

import requests

JAR_VERSION = "0.9.0"
JAR_FILE_NAME = f"sparql-anything-{JAR_VERSION}.jar"


def has_jar() -> bool:
    """Check the jar has been successfully downloaded in the installation folder."""
    files = os.listdir(get_module_path())
    return any(".jar" in file for file in files)


def get_jar_artifact_uri() -> str:
    """Retrieve the download url for latest SPARQL Anything release."""
    return f"https://download.eccenca.com/cmem-plugin-sparql-anything/sparql-anything-{JAR_VERSION}.jar"


def download_sparql_anything() -> None:
    """Download the passed version of the SPARQL Anything jar."""
    path2jar = Path(get_module_path()) / JAR_FILE_NAME
    dl_link = get_jar_artifact_uri()
    request = requests.get(dl_link, stream=True, timeout=10.0)
    request.raise_for_status()
    with (
        path2jar.open("wb") as jar,
    ):
        for data in request.iter_content(chunk_size=1024):
            jar.write(data)


def get_module_path() -> str:
    """Return the path to the PySPARQL Anything installation folder."""
    return os.path.realpath(Path(__file__).parent)


def get_path2jar() -> str:
    """Return the path to the PySPARQL Anything jar currently in use in the installation folder."""
    return str(Path(get_module_path()) / JAR_FILE_NAME)



"""Random values workflow plugin module"""

import io
import shlex
import tempfile
from collections.abc import Sequence
from subprocess import CompletedProcess, run

from cmem.cmempy.dp.proxy.graph import post_streamed
from cmem.cmempy.workspace.projects.resources.resource import get_resource_response
from cmem_plugin_base.dataintegration.context import ExecutionContext, ExecutionReport
from cmem_plugin_base.dataintegration.description import Icon, Plugin, PluginParameter
from cmem_plugin_base.dataintegration.entity import (
    Entities,
)
from cmem_plugin_base.dataintegration.parameter.code import SparqlCode
from cmem_plugin_base.dataintegration.parameter.graph import GraphParameterType
from cmem_plugin_base.dataintegration.parameter.resource import ResourceParameterType
from cmem_plugin_base.dataintegration.plugins import WorkflowPlugin
from cmem_plugin_base.dataintegration.types import BoolParameterType
from cmem_plugin_base.dataintegration.utils import setup_cmempy_user_access

from cmem_plugin_sparql_anything.constants import (
    DEFAULT_SPARQL,
    DOCUMENTATION,
    POLICY_TEMPLATE,
    QUERY_PARAMETER_DESCRIPTION,
    SPARQL_ANYTHING_ERROR_PATTERN,
)
from cmem_plugin_sparql_anything.utils import get_path2jar


@Plugin(
    label="SPARQL Anything",
    plugin_id="cmem_plugin_sparql_anything",
    description="Query anything with SPARQL to construct Knowledge Graphs.",
    documentation=DOCUMENTATION,
    icon=Icon(file_name="logo.svg", package=__package__),
    parameters=[
        PluginParameter(
            name="resource",
            label="File",
            description="Which resource file do you run query on? "
            "The dropdown shows file resources from the current project.",
            param_type=ResourceParameterType(),
        ),
        PluginParameter(name="query", label="Query", description=QUERY_PARAMETER_DESCRIPTION),
        PluginParameter(
            name="graph",
            label="Graph",
            description="Graph in which query result is stored.",
            param_type=GraphParameterType(allow_only_autocompleted_values=False),
        ),
        PluginParameter(
            name="replace_graph",
            label="Replace Graph",
            description="Enabling this option to replace graph triples.",
            param_type=BoolParameterType(),
        ),
    ],
)
class SPARQLAnything(WorkflowPlugin):
    """SPARQL Anything Workflow Plugin: Query file to generate knowledge graph"""

    def __init__(
        self,
        resource: str,
        graph: str,
        replace_graph: bool = False,
        query: SparqlCode = DEFAULT_SPARQL,
    ) -> None:
        self.resource = resource
        self.query = str(query)
        self.graph = graph
        self.replace_graph = replace_graph

    def execute(
        self,
        inputs: Sequence[Entities],  # noqa: ARG002
        context: ExecutionContext,
    ) -> None:
        """Run the workflow operator."""
        self.log.info("Start querying resource")
        setup_cmempy_user_access(context.user)

        with tempfile.NamedTemporaryFile(delete=True, suffix=self.resource) as resource_file:
            self._download_resource(context.task.project_id(), self.resource, resource_file)  # type: ignore[arg-type]
            self.post_result_to_graph(data=self._run_query(resource_file.name))
            context.report.update(ExecutionReport(entity_count=1, operation_desc="graph updated"))

    def _download_resource(self, project_id: str, resource: str, file: io.StringIO) -> None:
        """Download the resource and writes it to the temporary file."""
        self.log.info("Downloading resource")
        with get_resource_response(project_id, resource) as response:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)
        file.flush()

    def post_result_to_graph(self, data: bytes) -> None:
        """Post result to graph"""
        self.log.info(data.decode())
        post_streamed(graph=self.graph, file=io.BytesIO(data), replace=self.replace_graph)

    def _run_query(self, resource: str) -> bytes:
        """Run the SPARQL Anything jar with the provided query and resource."""
        self.log.info("Start SPARQL Anything")
        with (
            tempfile.NamedTemporaryFile(suffix=".sparql", delete=True) as query_file,
            tempfile.NamedTemporaryFile(suffix="anything.policy", delete=True) as policy_file,
        ):
            # Write policy content with proper resource file path
            policy_file.write(POLICY_TEMPLATE.format(query_file.name, resource).encode("utf-8"))
            policy_file.flush()

            # Replace resource placeholder in query with actual file path
            query_file.write(
                self.query.replace("{{resource_file}}", f"file://{resource}").encode("utf-8")
            )
            query_file.flush()

            # Build command with policy and query file paths
            cmd = (
                f"java -Djava.security.manager -Djava.security.policy={policy_file.name}"
                f" -jar {get_path2jar()} -q {query_file.name}"
            )
            output: CompletedProcess = run(shlex.split(cmd), check=False, capture_output=True)  # noqa: S603
        if SPARQL_ANYTHING_ERROR_PATTERN in output.stderr.decode("utf-8"):
            error = output.stderr.decode("utf-8").partition(SPARQL_ANYTHING_ERROR_PATTERN)[2]
            raise ValueError(f"{error}")

        return output.stdout  # type: ignore[no-any-return]
